local AddonName, GGL = ...

local assert, type, next, tonumber, tostring =
	  assert, type, next, tonumber, tostring
local pairs, ipairs, unpack = pairs, ipairs, unpack
local str_find, str_gsub, str_match, str_split, str_sub, format =
	  string.find, string.gsub, string.match, string.split, string.sub, format
local tblremove = table.remove

-- Embeds
LibStub("AceSerializer-3.0"):Embed(GGL)
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")
local LDBIcon = LibStub("LibDBIcon-1.0")

local Broker_GGL

local L = GGL.Locale

local InitEventFrame = CreateFrame("FRAME")
InitEventFrame:RegisterEvent("ADDON_LOADED")
InitEventFrame:RegisterEvent("PLAYER_LOGIN")
local InitFuncRegister = {}
local GGL_Init = false
local PlayerIsLogin = false

local function RunInitFunctions(event, addonName)
	if not event then
		RunInitFunctions("ADDON_LOADED", AddonName)
		RunInitFunctions("PLAYER_LOGIN")
		return
	end

	local addonLoadedState = false
	if addonName then
		local loaded, finished = IsAddOnLoaded(addonName)
		addonLoadedState = loaded and finished
	end

	if (event == "ADDON_LOADED" and addonLoadedState) or (event == "PLAYER_LOGIN" and PlayerIsLogin) then
		local tab = addonName and InitFuncRegister[event][addonName] or InitFuncRegister[event]
		if tab then
			for i = 1, #tab do
				tab[i]()
				tab[i] = nil
			end
		end
	end
end

local function InitEventFrame_OnEvent(frame, event, arg1, ...)
	if event == "ADDON_LOADED" and InitFuncRegister[event][arg1] then
		GGL_Init = GGL_Init or GGL:OnInitialize()
		RunInitFunctions(event, arg1)
	elseif event == "PLAYER_LOGIN" then
		PlayerIsLogin = true
		-- delay the first run, this "fix" some strange bugs in the loading process
		C_Timer.After(1, function() RunInitFunctions(event) end)
		InitEventFrame:UnregisterEvent(event)
	end
end
InitEventFrame:SetScript("OnEvent", InitEventFrame_OnEvent)

--- Adds a function that gets called with the ADDON_LOADED event
-- Main OnInitialize is run before this
-- call order is add order, when event already fired the function gets called immediately
-- @param		func		<function>
-- @param		addonName	<string> load with addon
function GGL:AddInitFunc(func, addonName)
	assert(type(func) == "function", "'func' must be a function.")

	addonName = addonName or AddonName

	if not InitFuncRegister["ADDON_LOADED"] then
		InitFuncRegister["ADDON_LOADED"] = {}
	end

	if not InitFuncRegister["ADDON_LOADED"][addonName] then
		InitFuncRegister["ADDON_LOADED"][addonName] = {}
	end

	InitFuncRegister["ADDON_LOADED"][addonName][ #InitFuncRegister["ADDON_LOADED"][addonName] + 1 ] = func

	RunInitFunctions("ADDON_LOADED", addonName)
end

--- Adds a function that gets called with the PLAYER_LOGIN event
-- call order is add order, when event already fired the function gets called immediately
-- @param		func		<function>
function GGL:AddInitFuncPlayerLogin(func)
	assert(type(func) == "function", "'func' must be a function.")

	if not InitFuncRegister["PLAYER_LOGIN"] then
		InitFuncRegister["PLAYER_LOGIN"] = {}
	end

	InitFuncRegister["PLAYER_LOGIN"][ #InitFuncRegister["PLAYER_LOGIN"]  + 1 ] = func

	RunInitFunctions("PLAYER_LOGIN")
end

function GGL:OnInitialize()
	if GoldGoblinLootDB then
		-- GGL.DB_VERSION
		if not GoldGoblinLootDB.dbVersion or GoldGoblinLootDB.dbVersion < GGL.DB_VERSION then
			wipe(GoldGoblinLootDB)
			if GoldGoblinLootSessionDB then
				wipe(GoldGoblinLootSessionDB)
			end
			-- set version
			GoldGoblinLootDB.dbVersion = GGL.DB_VERSION
		end
		-- GGL.DB_PLAYER_REGISTER_VERSION
		if not GoldGoblinLootDB.dbPlayerRegisterVersion or GoldGoblinLootDB.dbPlayerRegisterVersion < GGL.DB_PLAYER_REGISTER_VERSION then
			if GoldGoblinLootDB.global and GoldGoblinLootDB.global.PlayerRegister then
				wipe(GoldGoblinLootDB.global.PlayerRegister)
			end
			-- set version
			GoldGoblinLootDB.dbPlayerRegisterVersion = GGL.DB_PLAYER_REGISTER_VERSION
		end
		-- GGL.DB_SESSION_REGISTER_VERSION
		if not GoldGoblinLootDB.dbSessionRegisterVersion or GoldGoblinLootDB.dbSessionRegisterVersion < GGL.DB_SESSION_REGISTER_VERSION then
			if GoldGoblinLootSessionDB then
				wipe(GoldGoblinLootSessionDB)
			end
			-- set version
			GoldGoblinLootDB.dbSessionRegisterVersion = GGL.DB_SESSION_REGISTER_VERSION
		end
	end

	-- init db
	GGL.db = LibStub("AceDB-3.0"):New("GoldGoblinLootDB", GGL.dbdefaults, true).global
	--GGL.sessionDb = LibStub("AceDB-3.0"):New("GoldGoblinLootSessionDB", GGL.session_dbdefaults, true).global
	GoldGoblinLootSessionDB = GoldGoblinLootSessionDB or {}
	if GoldGoblinLootSessionDB.global then
		GoldGoblinLootSessionDB.profileKeys = nil
		local s = GoldGoblinLootSessionDB.global
		GoldGoblinLootSessionDB.global = nil
		GoldGoblinLootSessionDB = s
	end
	GGL.sessionDb = GoldGoblinLootSessionDB
	GGL.sessionDb.Sessions = GGL.sessionDb.Sessions or {}
	GGL.sessionDb.Archiv = GGL.sessionDb.Archiv or {}
	GGL.sessionDb.SessionRegister = GGL.sessionDb.SessionRegister or {}


	GGL:InstanceResets_Update()

	LDBIcon:Register("GoldGoblinLoot", Broker_GGL, GGL.db.MinimapButton)

	return true
end

GGL.session_dbdefaults = {
	global = {
		SessionRegister = {},
		RunningSession = nil,
		Sessions = {},
	},
}

function GGL:Print(msg)
	print("|cff33ff99GoldGoblinLoot|r: "..(msg or ""))
end

function GGL:FormatPrint(msg, ...)
	print("|cff33ff99GoldGoblinLoot|r: "..format(msg, ...))
end
-- ########################################
-- GGL event
-- ########################################
local EventFrame = CreateFrame("Frame")
local EventSaver = {}

local function EventFrame_OnEvent(self, event, ...)
	if GGL[EventSaver[event] or event] then
		GGL[EventSaver[event] or event](GGL, ...)
	end
end
EventFrame:SetScript("OnEvent", EventFrame_OnEvent)

function GGL:RegisterEvent(eventName, eventCall)
	if eventCall then
		EventSaver[eventName] = eventCall
	end
	EventFrame:RegisterEvent(eventName)
end

function GGL:UnregisterEvent(eventName)
	EventSaver[eventName] = nil
	EventFrame:UnregisterEvent(eventName)
end

-- ########################################
-- Callback Templates
-- ########################################
local CallbackProto = {
	["FireCallbackEvent"] = function(self, eventName, sender, ...)
		if not self.__callbacks or not self.__callbacks[eventName] or not self.__callbacks[eventName].exec then return end

		local event = self.__callbacks[eventName].exec
		local fix
		for i = 1, event.count do
			if event[i] then
				event[i](sender or self, ...)
			else
				fix = true
			end
		end
		if fix then
			local new = {}
			for i = 1, event.count do
				if event[i] then
					new[#new+1] = event[i]
				end
			end
			new.count = #new
			if self.__callbacks[eventName] then
				self.__callbacks[eventName].exec = new
			end
		end
	end,
	["RegisterCallbackEvent"] = function(self, eventName, method)
		if not eventName or not method then return end
		if not self.__callbacks then
			self.__callbacks = {}
		end
		if not self.__callbacks[eventName] then
			self.__callbacks[eventName] = {
				methods = {},
				exec = {
					count = 0
				},
			}
		end
		if not self.__callbacks[eventName].methods[method] then
			self.__callbacks[eventName].exec.count = self.__callbacks[eventName].exec.count + 1
			self.__callbacks[eventName].exec[self.__callbacks[eventName].exec.count] = method
			self.__callbacks[eventName].methods[method] = true
		end
	end,
	["UnregisterCallbackEvent"] = function(self, eventName, method)
		if not eventName or not method then return end
		if not self.__callbacks or not self.__callbacks[eventName] then return end

		local event = self.__callbacks[eventName]
		if event.methods[method] then
			local entrys = event.exec.count
			for i = 1, entrys do
				if event.exec[i] == method then
					event.exec[i] = nil
					event.methods[method] = nil
					break
				end
			end
		else
			return
		end
	end,
	["UnregisterAllCallbackEvents"] = function(self)
		if not self.__callbacks then return end
		wipe(self.__callbacks)
	end,
	["ReplaceCallbackMethod"] = function(self, eventName, oldMethod, newMethod)
        if self.__callbacks[eventName].methods[oldMethod] then
            for k,v in ipairs(self.__callbacks[eventName].exec) do
                if v == oldMethod then
                    self.__callbacks[eventName].exec[k] = newMethod
                    break
                end
            end
			self.__callbacks[eventName].methods[oldMethod] = nil
            self.__callbacks[eventName].methods[newMethod] = true
        end
	end,
}

function GGL:AddCallbackHandler(target)
	if not target or type(target) ~= "table" then
		return
	end
	for k,v in pairs(CallbackProto) do
		target[k] = v
	end
end
GGL:AddCallbackHandler(GGL)

-- ########################################
-- ItemInfoQuery
-- ########################################
local GetItemInfo = GetItemInfo
local ITEM_EVENT = "GET_ITEM_INFO_RECEIVED"
local ItemInfo_CallBacks = {}

local function RunItemInfoCallbacks(itemID)
	if itemID and ItemInfo_CallBacks[itemID] then
		for k, v in ipairs(ItemInfo_CallBacks[itemID]) do
			v[1](unpack(v[2]))
		end
		ItemInfo_CallBacks[itemID] = nil
	end
end

local ItemInfoFrame = CreateFrame("Frame")
ItemInfoFrame:RegisterEvent(ITEM_EVENT)

local function ItemInfoFrame_OnEvent(self, event, itemID, success)
	if itemID and ItemInfo_CallBacks[itemID] then
		if success == true then
			RunItemInfoCallbacks(itemID)
		elseif success == false then
			GGL:Print("Cant query item with ID "..itemID)
			ItemInfo_CallBacks[itemID] = nil
		else
			GGL:Print("Cant find item with ID "..itemID)
			ItemInfo_CallBacks[itemID] = nil
		end
	end
end
ItemInfoFrame:SetScript("OnEvent", ItemInfoFrame_OnEvent)

--- Run a callback when itemInfo is cached
-- @param	itemID		<number>	itemID
-- @param	callback	<func>		callback(...)
-- @param	...			<...>		callback(...)
function GGL:IsItemCached(itemID, callback, ...)
	local itemName = GetItemInfo(itemID)
	if not itemName then
		if not ItemInfo_CallBacks[itemID] then
			ItemInfo_CallBacks[itemID] = {}
		end
		ItemInfo_CallBacks[itemID][#ItemInfo_CallBacks[itemID] + 1] = { callback, { ... } }
	else
		callback(...)
	end
end

-- ########################################
-- LDB
-- ########################################
local LeftButtonRegister = {}

local TT_H_1, TT_H_2 = "|cff00FF00GoldGoblinLoot|r", string.format("|cffFFFFFF%s|r", GGL.ADDON_VERSION or "")
local TT_ENTRY = "|cFFCFCFCF%s:|r %s" --|cffFFFFFF%s|r"

local function ShowFrames_LeftButton()
	for func, funcSelf in pairs(LeftButtonRegister) do
		func(funcSelf)
	end
end

Broker_GGL = LDB:NewDataObject("GoldGoblinLoot", {
	type = "launcher",
	text = "GoldGoblinLoot",
	icon = GGL.MEDIA_PATH.."minimap", --617218,
	OnClick = function(self, button)
		if button == 'LeftButton' then
			--IsShiftKeyDown()
			if IsAltKeyDown() then
				GGL.GUI:ResetAllControlFrames()
			else
				ShowFrames_LeftButton()
			end
		elseif button == "RightButton" then
			GGL.Options:Open()
		elseif(button == 'MiddleButton') then

		else

		end
	end,
	OnTooltipShow = function(tooltip)
		tooltip:AddDoubleLine(TT_H_1, TT_H_2)
		tooltip:AddLine(format(TT_ENTRY, L["Right Click"], L["Open Options."]))
		tooltip:AddLine(format(TT_ENTRY, L["Left Click"], L["Open all session windows."]))
		tooltip:AddLine(format(TT_ENTRY, L["ALT + Left Click"], L["Reset all windows."]))
	end,
})

function GGL:MMB_SetShownState(state)
	GGL.db.MinimapButton.hide = not state
	if state then
		LDBIcon:Show("GoldGoblinLoot")
	else
		LDBIcon:Hide("GoldGoblinLoot")
	end
end

function GGL:RunMMFunctions()
	ShowFrames_LeftButton()
end

function GGL:RegisterMMFunction(funcSelf, method)
	if not method or type(method) ~= "function" then return end
	LeftButtonRegister[method] = funcSelf
end

function GGL:UnregisterMMFunction(method)
	if not method then return end
	LeftButtonRegister[method] = nil
end

-- ########################################
-- Cache table
-- ########################################

--- Create a table with weak key and value
-- @param		indexMethod		<function> __index
function GGL.GetCacheTable(indexMethod)
	local t = {}

	setmetatable(t, {
		-- weak key and value
		__mode = "kv",
		__index = indexMethod
	})

	return t
end

-- ########################################
-- Sync time offset
-- ########################################

local TIME_ONE_HOURE = 60 * 60
local TIME_CHECK = TIME_ONE_HOURE - (60*10)

function GGL.GetTimeOffset(syncTime)
	if not syncTime or syncTime <= 0 then
		return 0
	end
	local currentTime = GGL.GetTime()
	local offset = 0

	if (currentTime - syncTime) > TIME_CHECK then
		for i = 1, 24 do
			offset = offset + TIME_ONE_HOURE
			if currentTime - (syncTime+offset) < TIME_CHECK then
				break
			end
		end
	elseif (syncTime - currentTime) > TIME_CHECK then
		for i = 1, 24 do
			offset = offset + TIME_ONE_HOURE
			if syncTime - (currentTime+offset) < TIME_CHECK then
				offset = -offset
				break
			end
		end
	end

	return offset
end

-- ########################################
-- Money String (SharedXML/Util.lua - GetMoneyString)
-- ########################################
local BreakUpLargeNumbers = _G.BreakUpLargeNumbers

local COPPER_PER_GOLD, COPPER_PER_SILVER = _G.COPPER_PER_GOLD, _G.COPPER_PER_SILVER
local GOLD_TEXT, SILVER_TEXT, COPPER_TEXT = "|cffffd70ag|r", "|cffc7c7cfs|r", "|cffeda55fc|r"
local GOLD_ICON, SILVER_ICON, COPPER_ICON = "|T237618:%d::1|t", "|T237620:%d::1|t", "|T237617:%d::1|t"
--only text
local MONEY_GSC_FORMAT_SHORT 	= "%s"..GOLD_TEXT.." %02d"..SILVER_TEXT.." ".."%02d"..COPPER_TEXT
local MONEY_SC_FORMAT_SHORT 	= "%d"..SILVER_TEXT.." %02d"..COPPER_TEXT
local MONEY_C_FORMAT_SHORT 		= "%d"..COPPER_TEXT
 -- text with icon
local MONEY_GSC_FORMAT_SHORT_ICON 	= "%s"..GOLD_ICON.." %02d"..SILVER_ICON.." ".."%02d"..COPPER_ICON
local MONEY_SC_FORMAT_SHORT_ICON 	= "%d"..SILVER_ICON.." %02d"..COPPER_ICON
local MONEY_C_FORMAT_SHORT_ICON 	= "%d"..COPPER_ICON
-- colored text with icon
local MONEY_GSC_FORMAT_SHORT_ICON_COLORT 	= "|cffffd70a%s|r"..GOLD_ICON.." |cffc7c7cf%02d|r"..SILVER_ICON.." ".."|cffeda55f%02d|r"..COPPER_ICON
local MONEY_SC_FORMAT_SHORT_ICON_COLORT 	= "|cffc7c7cf%d|r"..SILVER_ICON.." |cffeda55f%02d|r"..COPPER_ICON
local MONEY_C_FORMAT_SHORT_ICON_COLORT	 	= "|cffeda55f%d|r"..COPPER_ICON

function GGL.GetMoneyText(money)
	money = tonumber(money) or 0

    local isNegative = money < 0
    money = abs(money)

    local gold = floor(money / COPPER_PER_GOLD)
    local silver = floor((money % COPPER_PER_GOLD) / COPPER_PER_SILVER)
    local copper = floor(money % COPPER_PER_SILVER)
    local text

    if money == 0 then
        text = format(MONEY_C_FORMAT_SHORT, 0)
    else
        if gold > 0 then
            text = format(MONEY_GSC_FORMAT_SHORT, (BreakUpLargeNumbers(gold) or "0"), silver or 0, copper or 0)
        elseif silver > 0 then
            text = format(MONEY_SC_FORMAT_SHORT, silver or 0, copper or 0)
        else
            text = format(MONEY_C_FORMAT_SHORT, copper or 0)
        end

        if isNegative then
            text = "|cffff0000-"..text.."|r"
        end
    end
    return text
end

function GGL.GetMoneyTextShort(money)
	money = tonumber(money) or 0

    local isNegative = money < 0
    money = abs(money)

    local gold = floor(money / COPPER_PER_GOLD)
    local silver = floor((money % COPPER_PER_GOLD) / COPPER_PER_SILVER)
    local copper = floor(money % COPPER_PER_SILVER)
    local text

    if money == 0 then
        text = format(MONEY_C_FORMAT_SHORT, 0)
    else
        if gold > 0 then
			text = (BreakUpLargeNumbers(gold) or 0)..GOLD_TEXT
		elseif silver > 0 then
			text = silver..SILVER_TEXT
		else
			text = copper..COPPER_TEXT
		end

        if isNegative then
            text = "|cffff0000-"..text.."|r"
        end
    end
    return text
end

function GGL.GetMoneyTextWithIcons(money, fontHeight, coloredText)
	fontHeight = fontHeight or 14
	fontHeight = fontHeight - 2
	money = tonumber(money) or 0

    local isNegative = money < 0
    money = abs(money)

    local gold = floor(money / COPPER_PER_GOLD)
    local silver = floor((money % COPPER_PER_GOLD) / COPPER_PER_SILVER)
    local copper = floor(money % COPPER_PER_SILVER)
    local text

    if money == 0 then
        text = format(MONEY_C_FORMAT_SHORT_ICON, 0, fontHeight)
    else
        if gold > 0 then
            text = format(coloredText and MONEY_GSC_FORMAT_SHORT_ICON_COLORT or MONEY_GSC_FORMAT_SHORT_ICON, (BreakUpLargeNumbers(gold) or "0"), fontHeight, silver or 0, fontHeight, copper or 0, fontHeight)
        elseif silver > 0 then
            text = format(coloredText and MONEY_SC_FORMAT_SHORT_ICON_COLORT or MONEY_SC_FORMAT_SHORT_ICON, silver or 0, fontHeight, copper or 0, fontHeight)
        else
            text = format(coloredText and MONEY_C_FORMAT_SHORT_ICON_COLORT or MONEY_C_FORMAT_SHORT_ICON, copper or 0, fontHeight)
        end

        if isNegative then
            text = "|cffff0000-"..text.."|r"
        end
    end
    return text
end
-- ########################################
-- Player Position
-- Events
--		"GGL_OnMapChange"	(mapID, mapName, mapTypeFlag)
-- ########################################
local GetMapInfo, GetBestMapForUnit = C_Map.GetMapInfo, C_Map.GetBestMapForUnit
local lastMapID = 0

-- /dump C_Map.GetBestMapForUnit("player"), EJ_GetInstanceForMap(C_Map.GetBestMapForUnit("player"))
function GGL:GetZoneData(mapID)
	--GetMapInfo
	mapID = mapID or GetBestMapForUnit("player")
	mapID = GGL.MAP_BASE_ID[mapID] or mapID
	if not mapID then return end

	-- basic map data
	local mapData = GetMapInfo(mapID)
	local instanceID = GGL.FIX_MAPID_TO_INSTANCID[mapID] or EJ_GetInstanceForMap(mapID)
	local _, _, difficultyID = GetInstanceInfo()

	-- check for mapType
	local mapTypeFlag

	if instanceID and difficultyID and (GGL.INSTANCE_ID_DATA[instanceID] or GGL.INSTANCE_ID_DATA[instanceID..":"..difficultyID]) then
		mapTypeFlag = GGL.DIFFICULYID_TO_MAP_TYPE_FLAG[difficultyID or 0]
		mapID = GGL.INSTANCE_ID_DATA[instanceID..":"..difficultyID] and GGL.INSTANCE_ID_DATA[instanceID..":"..difficultyID][2] or GGL.INSTANCE_ID_DATA[instanceID][2]
	elseif GGL.MAP_ID_DATA[mapID] and GGL.MAP_ID_DATA[mapID][4] then
		mapTypeFlag = GGL.MAP_ID_DATA[mapID][4]
	end

	return mapID, mapData and mapData.name or nil, mapTypeFlag
end

function GGL:GetCurrentMapID()
	return lastMapID
end

function GGL:ZONE_CHANGED_NEW_AREA()
	local mapID, mapName, mapTypeFlag = GGL:GetZoneData()
	if mapID == lastMapID then return end
	-- Fire
	lastMapID = mapID
	if mapID then
		GGL:FireCallbackEvent("GGL_OnMapChange", nil, mapID, mapName, mapTypeFlag)
	end
end
GGL:RegisterEvent("ZONE_CHANGED_NEW_AREA")
GGL:RegisterEvent("PLAYER_ENTERING_WORLD", "ZONE_CHANGED_NEW_AREA")
GGL:RegisterEvent("ZONE_CHANGED", "ZONE_CHANGED_NEW_AREA")
GGL:RegisterEvent("ZONE_CHANGED_INDOORS", "ZONE_CHANGED_NEW_AREA")
GGL:RegisterEvent("NEW_WMO_CHUNK", "ZONE_CHANGED_NEW_AREA")
