local AddonName, GGL = ...

local PriceSource = {}
GGL.PriceSource = PriceSource

local L = GGL.Locale

-- lua
local ipairs = ipairs
local format = string.format

-- wow
local GetItemInfo = GetItemInfo
local GetPetInfoByItemID = C_PetJournal.GetPetInfoByItemID

-- ##################
-- local
-- ##################
local SOURCE_LIST = {}
local SORTED_SOURCE_LIST = {}
local SOURCE_LISTS_SHORT = { short = {}, long = {} }
local SOURCE_LIST_NAME = {}

local DEFAULT_SOURCE_ID = "d"

-- ##################
-- main
-- ##################

--- Add a new Price source
-- @param   id              <string>    Unique price source ID
-- @param   displayName     <string>    Displayname
-- @param   method          <function>  (itemID) -> @return copperPrice
function PriceSource:RegisterSource(id, displayName, method)
    if not SOURCE_LIST[id] and not SOURCE_LIST_NAME[displayName] then
        SOURCE_LIST_NAME[id] = displayName
        SOURCE_LIST[id] = method

        SORTED_SOURCE_LIST[#SORTED_SOURCE_LIST+1] = id
    end
end

function PriceSource:GetFunction(id)
    return SOURCE_LIST[id or DEFAULT_SOURCE_ID] or SOURCE_LIST[DEFAULT_SOURCE_ID]
end

function PriceSource:GetDropDownListData()
    local list = {}
    for k,v in ipairs(SORTED_SOURCE_LIST) do
        list[v] = SOURCE_LIST_NAME[v]
    end
    return list, SORTED_SOURCE_LIST
end

function PriceSource:GetDefaultID()
    return DEFAULT_SOURCE_ID
end

function PriceSource:GetName(priceSourceID)
    priceSourceID = priceSourceID or DEFAULT_SOURCE_ID
    return SOURCE_LIST_NAME[priceSourceID]
end

function PriceSource:IsValidID(priceSourceID)
    return SOURCE_LIST_NAME[priceSourceID or DEFAULT_SOURCE_ID] and true or false
end

-- ##################
-- Blizzard
-- ##################
PriceSource:RegisterSource(DEFAULT_SOURCE_ID, L["Vendor price"], function(itemID, itemString)
    if not itemID then return 0 end
    local _, _, _, _, _, _, _, _, _, _, baseItemSellPrice = GetItemInfo(itemID)
    _, _, _, _, _, _, _, _, _, _, itemString = GetItemInfo(itemString or itemID)
    return baseItemSellPrice or 0, itemString or 0
end)

-- ##################
-- TSM
-- ##################
local function LoadTSMSources()
    if not _G.TSM_API then return end
    local TSM_API = _G.TSM_API
    local PET_FORMAT = "p:%d:1:3"

    local function GetTSMItem(itemID, itemString)
        local _, _, _, _, _, itemClassID, itemSubClassID = GetItemInfoInstant(itemID)
        if itemClassID == 15 and itemSubClassID == 2 then
            -- it should be a battle pet
            local _, _, _, _, _, _, _, _, _, _, _, _, speciesID = GetPetInfoByItemID(itemID)
            if speciesID then
                return format(PET_FORMAT, speciesID)
            end
        end

        if itemString then
            return TSM_API.ToItemString(itemString) or "i:"..itemID
        else
            return "i:"..itemID
        end
    end

    -- the key is TSM:<key> -> TSM:DBRegionMarketAvg
    local USED_SOURCES = {
        --["Custom"]                  = "Custom Price Source",
        ["DBHistorical"]            = {"TSM: Historical Price",             "t1"},
        ["DBMarket"]                = {"TSM: Market Value",                 "t2"},
        ["DBMinBuyout"]             = {"TSM: Min Buyout",                   "t3"},
        ["DBRegionHistorical"]      = {"TSM: Region Historical Price",      "t4"},
        ["DBRegionMarketAvg"]       = {"TSM: Region Market Value Avg",      "t5"},
        ["DBRegionMinBuyoutAvg"]    = {"TSM: Region Min Buyout Avg",        "t6"},
        ["DBRegionSaleAvg"]         = {"TSM: Region Global Sale Average",   "t7"},
        --["VendorSell"]              = {"TSM: VendorSell",                   "t8"},
    }
    local tmpTsmPriceSourceKeys = {}
    TSM_API.GetPriceSourceKeys(tmpTsmPriceSourceKeys)

    for i = 1, #tmpTsmPriceSourceKeys do
        if USED_SOURCES[tmpTsmPriceSourceKeys[i]] then
            local sourceKey = tmpTsmPriceSourceKeys[i]
            PriceSource:RegisterSource(USED_SOURCES[sourceKey][2], USED_SOURCES[sourceKey][1], function(itemID, itemString)
                if not itemID then return 0 end
                if itemString then
                    return TSM_API.GetCustomPriceValue(sourceKey, GetTSMItem(itemID)) or 0, TSM_API.GetCustomPriceValue(sourceKey, GetTSMItem(itemID, itemString))
                else
                    return TSM_API.GetCustomPriceValue(sourceKey, GetTSMItem(itemID)) or 0
                end

            end)
        end
    end

end

GGL:AddInitFuncPlayerLogin(LoadTSMSources)

-- ##################
-- TheUndermineJournal
-- ##################
local function LoadTUJSources()
    if not _G.TUJMarketInfo then return end
    local TUJMarketInfo = _G.TUJMarketInfo

    local USED_SOURCES = {
        ["recent"]            = {"TUJ: 3-Day Price",                "u1"},
        ["market"]            = {"TUJ: 14-Day Price",               "u2"},
        ["stddev"]            = {"TUJ: 14-Day Std Dev",             "u3"},
        ["globalMedian"]      = {"TUJ: Global Median",              "u4"},
        ["globalMean"]        = {"TUJ: Global Mean",                "u5"},
        ["globalStdDev"]      = {"TUJ: Global Std Dev",             "u6"},
    }

    local SortOrder = {"recent", "market", "stddev", "globalMedian", "globalMean", "globalStdDev"}

    for i = 1, #SortOrder do
        if USED_SOURCES[SortOrder[i]] then
            local sourceKey = SortOrder[i]

            PriceSource:RegisterSource(USED_SOURCES[sourceKey][2], USED_SOURCES[sourceKey][1], function(itemID, itemString)
                if not itemID then return 0 end
                itemID = TUJMarketInfo(itemID)
                if itemID and itemID[sourceKey] then
                    return itemID[sourceKey]
                end
                return 0
            end)

        end
    end
end

GGL:AddInitFuncPlayerLogin(LoadTUJSources)

-- ##################
-- Auctionator
-- ##################

local function LoadAuctionatorSources()
    if not _G.Auctionator then return end

    local GetByItemID = _G.Auctionator.API.v1.GetAuctionPriceByItemID
    local GetByItemLink = _G.Auctionator.API.v1.GetAuctionPriceByItemLink

    _G.Auctionator.API.InternalVerifyID(AddonName)

    PriceSource:RegisterSource("ac", "Auctionator", function(itemID, itemString)
        if not itemID then return 0 end
        local baseItemSellPrice = GetByItemID(AddonName, itemID) or 0
        return baseItemSellPrice, itemString and GetByItemLink(AddonName, itemString) or baseItemSellPrice
    end)
end

GGL:AddInitFuncPlayerLogin(LoadAuctionatorSources)