local AddonName, GGL = ...

local L = GGL.Locale

local Statistics = {}
GGL.Statistics = Statistics

-- lua
local setmetatable = setmetatable
local tbl_insert = table.insert
local str_sub = string.sub

-- wow
local GetItemInfo = GetItemInfo

-- GGL
local MAIN_SOURCE_FLAG, SUB_SOURCE_FLAG = GGL.MAIN_SOURCE_FLAG, GGL.SUB_SOURCE_FLAG

-- ##################
-- local
-- ##################
local function ProcessTopItems(maxItems, itemList, newItem, itemString, newItemValue)
    if not itemList.items then itemList.items = {} end
    if itemList.items[itemString or newItem] then
        return
    end
    local itemKey = itemString or newItem
    for i = 1, maxItems do
        if not itemList[i] then
            -- list is no full so just add it
            itemList[i] = {itemKey, newItemValue}
            itemList.lastUpdate = GGL.GetTime()
            itemList.items[itemKey] = true
            return
        elseif newItemValue > itemList[i][2] then
            -- better item found
            tbl_insert(itemList, i, {itemKey, newItemValue})
            -- remove outmoved item
            itemList[maxItems+1] = nil
            itemList.lastUpdate = GGL.GetTime()
            itemList.items[itemKey] = true
            return
        end
    end
end

local function MergeTopItemsTable(maxItems, destList, srcList)
    if not destList.items then destList.items = {} end
    local update
    for srcI = 1, maxItems do
        local item = srcList[srcI]
        if item and not destList.items[item[1]] then
            for destI = 1, maxItems do
                if not destList[destI] then
                    destList[destI] = item
                    destList.items[item[1]] = true
                    update = true
                    break
                elseif item[2] > destList[destI][2] then
                    tbl_insert(destList, destI, item)
                    destList.items[item[1]] = true
                    update = true
                    break
                end
            end
        end
    end

    --cleanup
    if #destList > maxItems then
        for i = #destList, maxItems, -1 do
            destList[i] = nil
        end
    end
    if update then
        destList.lastUpdate = GGL.GetTime()
    end
end

local function WipeStatisticsTables(tab)
    for k,v in pairs(tab) do
        if type(v) == "table" then
            wipe(tab[k])
        else
            tab[k] = 0
        end
    end
end

-- ##################
-- Base
-- ##################
local NUM_TOP_ITEMS_OVERALL = GGL.MAX_TOP_ITEMS -- Constants.lua
local NUM_TOP_ITEMS_PLAYER = 10

local SessionRegister = {}

-- ################
-- Callbacks
local function CB_OnClose(session)
    if not SessionRegister[session] then return end
    Statistics:SealSession(session)
end

local function CB_SetDataRefresh(session)
    if SessionRegister[session] then
        SessionRegister[session]:ResetSession()
    end
end

local function CB_CFrameOnNewItem(session, itemID, itemString, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
    if SessionRegister[session] then
        SessionRegister[session]:AddItem(itemID, session:LoggedItemUseVariations(itemID, itemString) and itemString or nil, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
    end
end

local function CB_CFrameOnAddMoney(session, copper, sourcePlayer)
    if SessionRegister[session] then
        SessionRegister[session]:AddRawGold(copper, sourcePlayer)
    end
end
-- Callbacks end
-- ################

local STATISTICS_PROTO = {
    --- Create needed session data
    ["Constructor"] = function(self)
        self.sessionDB = self.session.db
        self.overview = {
            itemPrices = {},
            itemsData = {},
            itemDataCache = {},
            itemQualityCount = {},
            itemQualityValue = {},
            itemCount = {},
            topItems = {},
            numItems = 0,
            numUniqueItems = 0,
            fullItemValue = 0,
        }
        self.players = {}
        self.iPlayers = {}
        self.mergeID = {
            topItems = {},
        }

        for quality = 0, #GGL.ITEM_QUALITY_INFO do
            self.overview.itemQualityCount[quality] = 0
            self.overview.itemQualityValue[quality] = 0
        end
    end,

    ["StartTracking"] = function(self)
        -- cb
        self.session:RegisterCallbackEvent("SessionTemplate_AddItem", CB_CFrameOnNewItem)
        self.session:RegisterCallbackEvent("SessionTemplate_AddMoney", CB_CFrameOnAddMoney)
        self.session:RegisterCallbackEvent("SessionTemplate_SetDataRefresh", CB_SetDataRefresh)
        self.session:RegisterCallbackEvent("SessionTemplate_SessionClosed", CB_OnClose)
    end,

    ["StopTracking"] = function(self)
        -- cb
        self.session:UnregisterCallbackEvent("SessionTemplate_AddItem", CB_CFrameOnNewItem)
        self.session:UnregisterCallbackEvent("SessionTemplate_AddMoney", CB_CFrameOnAddMoney)
        self.session:UnregisterCallbackEvent("SessionTemplate_SetDataRefresh", CB_SetDataRefresh)
        self.session:UnregisterCallbackEvent("SessionTemplate_SessionClosed", CB_OnClose)
    end,

    --- Seals a session so no more data can be added
    ["Seal"] = function(self)
        Statistics:SealSession(self.session)
    end,

    --- Resets all data from that session
    ["ResetSession"] = function(self)
        local session = self.session

        wipe(self.players)
        WipeStatisticsTables(self.overview)
        WipeStatisticsTables(self.mergeID)

        for quality = 0, #GGL.ITEM_QUALITY_INFO do
            self.overview.itemQualityCount[quality] = 0
            self.overview.itemQualityValue[quality] = 0
        end
    end,

    --- Adds a item
    ["AddItem"] = function(self, itemID, itemString, zoneID, sourceID, ammount, copperValue, mainSourceFlag, subSourceFlag, sourcePlayer)
        local player = self:GetPlayerTable(sourcePlayer)
        if not player then return end
        local overview = self.overview

        ammount = ammount or 1

        local itemKey = itemString or itemID
        local itemName, itemLink, itemRarity
        if overview.itemDataCache[itemKey] and overview.itemDataCache[itemKey][2] then
            itemName, itemLink, itemRarity = overview.itemDataCache[itemKey][1], overview.itemDataCache[itemKey][2], overview.itemDataCache[itemKey][3]
        else
            itemName, itemLink, itemRarity = GetItemInfo(itemKey)
            overview.itemDataCache[itemKey] = {itemName, itemLink, itemRarity}
        end

        -- add rawData into log
        player.itemsData[#player.itemsData + 1] = {player, itemID, itemString, zoneID, sourceID, ammount, mainSourceFlag, subSourceFlag}
        overview.itemsData[#overview.itemsData + 1] = player.itemsData[#player.itemsData]

        -- increase item counter
        player.numItems = player.numItems + ammount
        overview.numItems = self.overview.numItems + ammount

        -- increase item counter by quality
        if player.itemQualityCount[itemRarity] then
            player.itemQualityCount[itemRarity] = player.itemQualityCount[itemRarity] + ammount
            overview.itemQualityCount[itemRarity] = overview.itemQualityCount[itemRarity] + ammount
        end

        -- increase item value
        player.fullItemValue = player.fullItemValue + (ammount * copperValue)
        overview.fullItemValue = overview.fullItemValue + (ammount * copperValue)

        -- increase item value by quality
        if player.itemQualityValue[itemRarity] then
            player.itemQualityValue[itemRarity] = player.itemQualityValue[itemRarity] + (ammount * copperValue)
            overview.itemQualityValue[itemRarity] = overview.itemQualityValue[itemRarity] + (ammount * copperValue)
        end

        -- check if item is added as uniqueItems
        if not player.itemCount[itemKey] then
            player.itemCount[itemKey] = 0
            player.numUniqueItems = player.numUniqueItems + 1
            ProcessTopItems(NUM_TOP_ITEMS_PLAYER, player.topItems, itemID, itemString, copperValue)
        end
        player.itemCount[itemKey] = player.itemCount[itemKey] + ammount

        -- check if item is added as uniqueItems (overview)
        if not overview.itemCount[itemKey] then
            overview.itemCount[itemKey] = 0
            overview.itemPrices[itemKey] = copperValue
            overview.numUniqueItems = overview.numUniqueItems + 1
            -- check topItems
            ProcessTopItems(NUM_TOP_ITEMS_OVERALL, overview.topItems, itemID, itemString, copperValue)
        end
        overview.itemCount[itemKey] = overview.itemCount[itemKey] + ammount
    end,

    --- Adds a item
    ["AddRawGold"] = function(self, copper, sourcePlayer)
        local player = self:GetPlayerTable(sourcePlayer)
        if not player then return end

        -- set rawGold
        player.rawGold = player.rawGold + copper
    end,

    --- Gets the overview data table
    ["GetOverviewDataTable"] = function(self)
        return self.overview
    end,

    --- Gets a playerTable
    ["GetPlayerTable"] = function(self, playerTemplate, noCreation)
        if not playerTemplate then return end
        if not self.players[playerTemplate.syncID] and not noCreation then
            self.players[playerTemplate.syncID] = {
                syncID = playerTemplate.syncID,
                itemsData = {},
                goldData = {},
                itemQualityCount = {},
                itemQualityValue = {},
                itemCount = {},
                topItems = {},

                -- values
                numItems = 0,
                fullItemValue = 0,
                rawGold = 0,
                numUniqueItems = 0
            }

            for quality = 0, #GGL.ITEM_QUALITY_INFO do
                self.players[playerTemplate.syncID].itemQualityCount[quality] = 0
                self.players[playerTemplate.syncID].itemQualityValue[quality] = 0
            end

            self.iPlayers[#self.iPlayers + 1] = self.players[playerTemplate.syncID]
        end
        return self.players[playerTemplate.syncID]
    end,

    --- Gets itemcount for a mergeID
    ["GetItemCountForMergeID"] = function(self, mergeID, itemID)
        if mergeID and type(mergeID) == "table" and mergeID.syncID then
            mergeID = self.sessionDB.mergeIDList[mergeID.syncID]
        end
        if not mergeID then return 0 end
        local count = 0

        for syncID, pMergeID in pairs(self.sessionDB.mergeIDList) do
            if mergeID == pMergeID and self.players[syncID] and self.players[syncID].itemCount[itemID] then
                count = count + self.players[syncID].itemCount[itemID]
            end
        end

        return count
    end,

    --- Gets rawgold for a mergeID
    ["GetRawGoldForMergeID"] = function(self, mergeID)
        if mergeID and type(mergeID) == "table" and mergeID.syncID then
            mergeID = self.sessionDB.mergeIDList[mergeID.syncID]
        end
        if not mergeID then return 0 end
        local count = 0

        for syncID, pMergeID in pairs(self.sessionDB.mergeIDList) do
            if mergeID == pMergeID and self.players[syncID] then
                count = count + self.players[syncID].rawGold
            end
        end

        return count
    end,

    --- Gets itemValueGold for a mergeID
    ["GetItemValueGoldForMergeID"] = function(self, mergeID)
        if mergeID and type(mergeID) == "table" and mergeID.syncID then
            mergeID = self.sessionDB.mergeIDList[mergeID.syncID]
        end
        if not mergeID then return 0 end
        local count = 0

        for syncID, pMergeID in pairs(self.sessionDB.mergeIDList) do
            if mergeID == pMergeID and self.players[syncID] then
                count = count + self.players[syncID].fullItemValue
            end
        end

        return count
    end,

    --- Gets number of chars in challenge for a mergeID
    ["GetNumCharsForMergeID"] = function(self, mergeID)
        if mergeID and type(mergeID) == "table" and mergeID.syncID then
            mergeID = self.sessionDB.mergeIDList[mergeID.syncID]
        end
        if not mergeID then return 0 end
        local count = 0

        for syncID, pMergeID in pairs(self.sessionDB.mergeIDList) do
            if mergeID == pMergeID and self.sessionDB.lootLog[syncID] and #self.sessionDB.lootLog[syncID] > 0 then
                count = count + 1
            end
        end

        return count
    end,

    --- Gets reset data for a char
    -- numEntered, numReset
    ["GetInstanceResetDataForMergeID"] = function(self, mergeID)
        if mergeID and type(mergeID) == "table" and mergeID.syncID then
            mergeID = self.sessionDB.mergeIDList[mergeID.syncID]
        end
        if not mergeID then return end

        if self.sessionDB.instanceResetData[mergeID] then
            local numReset = 0
            for i = 1, #self.sessionDB.instanceResetData[mergeID] do
                if str_sub(self.sessionDB.instanceResetData[mergeID][i], 1, 1) == "r" then
                    numReset = numReset + 1
                end
            end
            return #self.sessionDB.instanceResetData[mergeID], numReset
        else
            return
        end
    end,

    --- Gets topItemList for a mergeID
    ["GetTopListForMergeID"] = function(self, mergeID)
        if mergeID and type(mergeID) == "table" and mergeID.syncID then
            mergeID = self.sessionDB.mergeIDList[mergeID.syncID]
        end
        if not mergeID then return end

        local topEntry = self.mergeID.topItems[mergeID]
        local needsUpdate

        -- check all playerMergeID's if we need a update
        if topEntry then
            for syncID, pMergeID in pairs(self.sessionDB.mergeIDList) do
                if pMergeID == mergeID and self.players[syncID] then
                    if not self.players[syncID].topItems.lastUpdate or not topEntry.lastUpdate then
                        needsUpdate = true
                        break
                    elseif self.players[syncID].topItems.lastUpdate ~= topEntry.lastUpdate then
                        needsUpdate = true
                        break
                    end
                end
            end
        else
            self.mergeID.topItems[mergeID] = {}
            topEntry = self.mergeID.topItems[mergeID]
            needsUpdate = true
        end

        -- update needed
        if needsUpdate then
            for syncID, pMergeID in pairs(self.sessionDB.mergeIDList) do
                if pMergeID == mergeID and self.players[syncID] then
                    MergeTopItemsTable(NUM_TOP_ITEMS_PLAYER, topEntry, self.players[syncID].topItems)
                end
            end
        end

        return topEntry
    end,

    --- Gets topItemList for a mergeID
    ["GetActivityPercentForMergeID"] = function(self, mergeID)
        if mergeID and type(mergeID) == "table" and mergeID.syncID then
            mergeID = self.sessionDB.mergeIDList[mergeID.syncID]
        end
        if not mergeID then return end
        local mergeIDList = self.sessionDB.mergeIDList

        local endTime = self.sessionDB.endTime
        local curTime = GGL.GetTime()
        curTime = curTime > endTime and endTime or curTime
        local fullDuration = curTime - self.sessionDB.startTime

        local cuValue, lastValue = 0, nil
        for index, statusEntry in ipairs(self.sessionDB.activityLog) do
            if mergeIDList[statusEntry[1]] == mergeID then
                if not lastValue and statusEntry[3] == 1 then
                    lastValue = statusEntry[2]
                elseif statusEntry[3] == 2 then
                    lastValue = lastValue or self.sessionDB.startTime
                    cuValue = cuValue + (statusEntry[2] - lastValue)
                    lastValue = nil
                end
            end
        end

        if lastValue then
            cuValue = cuValue + (curTime - lastValue)
        elseif not lastValue and cuValue == 0 then
            return 100
        end

        return cuValue > 0 and ((100 / fullDuration) * cuValue) or 0
    end,

    --- Gets a toplist of all Players of the current session
    -- highest -> lowest { playerTemplate, fullGold }
    -- @param       combineMergeID      <bool> combine values by mergeID
    ["GetPlayerTopList"] = function(self, combineMergeID)
        -- Templates.Sync.GetPlayerFromSync(playerSyncID, true)
        local topList = {}
        local tmpPlayerRegister = {}

        for i = 1, #self.iPlayers do
            local player = self.iPlayers[i]
            local playerKey = player.syncID
            if combineMergeID and self.sessionDB.mergeIDList[playerKey] then
                playerKey = self.sessionDB.mergeIDList[playerKey]
            end

            if not tmpPlayerRegister[playerKey] then
                -- local playerTemplate = GGL.Templates.Sync.GetPlayerFromSync(playerSyncID, true)
                topList[#topList+1] = {

                }
            end
        end
    end,

    --- Get npc data for all or single player
    -- @param       playerTemplate          <template:playerTemplate> if nil overall data is used
    -- @param       useMergeID              <bool> playerTemplate is a mergeID or use the mergeID from the playerTemplate
    -- @return npcData = {
    --     [npcID] = {
    --         itemsByQuality = {}, -- [qualityID] = numItems
    --         itemCount = {count = 0}, -- [itemID] = numItems (count == summary)
    --         fullItemValue = 0, -- full item value in copper
    --         fullItemCount = 0, -- sum of items
    --     }
    -- }, npcCache -- [npcID] = npcName
    ["GetNPCData"] = function(self, playerTemplate, useMergeID)
        local itemsData
        if not playerTemplate then
            -- no playerTemplate use overall data
            itemsData = self.overview.itemsData
        elseif playerTemplate and type(playerTemplate) == "string" then
            -- playerTemplate is the mergeID
            useMergeID = playerTemplate
        elseif useMergeID and playerTemplate.mergeID then
            -- use mergeID so use overall data
            itemsData = self.overview.itemsData
            useMergeID = playerTemplate.mergeID
        else
            -- just use playerTemplate, we can use playerData
            local player = self:GetPlayerTable(playerTemplate)
            if not player or not player.itemsData then return end
            itemsData = player.itemsData
            useMergeID = nil
        end

        -- locals
        local npcData = {}
        local itemDataCache = self.overview.itemDataCache
        local npcCache = GoldGoblinLootDataDB and GoldGoblinLootDataDB["npc_names"] or {}

        -- get npc data
        for i = 1, #itemsData do
            local entry = itemsData[i]
            -- check for NPC source
            if entry[7] == MAIN_SOURCE_FLAG.NPC and
                (not playerTemplate or (playerTemplate and playerTemplate == entry[1]) ) and
                (not useMergeID or (useMergeID and playerTemplate.mergeID == useMergeID) ) then

                -- add new npc
                if not npcData[entry[5]] then
                    npcData[entry[5]] = {
                        itemsByQuality = {},
                        itemCount = {
                            count = 0 -- full count
                        },
                        fullItemValue = 0,
                        fullItemCount = 0,
                    }
                end

                local npcEntry = npcData[entry[5]]
                local itemKey = entry[3] or entry[2]
                local itemCache = itemDataCache[itemKey]
                if not itemCache or not itemCache[2] then
                    local itemName, itemLink, itemRarity = GetItemInfo(itemKey)
                    self.overview.itemDataCache[itemKey] = {itemName, itemLink, itemRarity}
                end
                -- increase quality counter
                npcEntry[ itemCache[3] ] = (npcEntry[ itemCache[3] ] or 0) + entry[6]

                -- set itemRegister
                if not npcEntry.itemCount[itemCache[2]] then
                    -- create new table for quality if needed
                    if not npcEntry.itemsByQuality[itemCache[3]] then npcEntry.itemsByQuality[itemCache[3]] = {} end
                    -- add into quolity itemList
                    npcEntry.itemsByQuality[itemCache[3]][ #npcEntry.itemsByQuality[itemCache[3]] + 1 ] = itemCache[2]
                    -- create counter value
                    npcEntry.itemCount[itemCache[2]] = 0
                end
                -- increase item counter
                npcEntry.itemCount[itemCache[2]] = npcEntry.itemCount[itemCache[2]] + entry[6]
                npcEntry.itemCount.count = npcEntry.itemCount.count + entry[6]
                npcEntry.fullItemCount = npcEntry.fullItemCount + entry[6]
                -- increase item value
                npcEntry.fullItemValue = npcEntry.fullItemValue + (entry[6] * self.overview.itemPrices[itemKey])
            end
        end

        return npcData, npcCache
    end,
}
local STATISTICS_PROTO_MT = {__index = STATISTICS_PROTO}

--- Register a session
-- This justs allows the session to add data
-- @param       session         <template:session>
function Statistics:RegisterSession(session)
    if SessionRegister[session] then return end
    SessionRegister[session] = {
        isSeald = false,
        session = session,
    }
    setmetatable(SessionRegister[session], STATISTICS_PROTO_MT)

    SessionRegister[session]:Constructor()

    return SessionRegister[session]
end

--- Seals a session so no more data can be added
-- @param       session         <template:session>
function Statistics:SealSession(session)
    if not SessionRegister[session] then return end
    SessionRegister[session].isSeald = true

    -- cb
    SessionRegister[session]:StopTracking()
end

--- Gets data from a session
-- @param       dataKey         <string> data to get
-- @param       ...             <var> STATISTICS_PRIVAT_PROTO function calls
function Statistics:GetData(session, dataKey, ...)
    if not SessionRegister[session] then return end

    if SessionRegister[session][dataKey] then
        return SessionRegister[session][dataKey](SessionRegister[session], ...)
    end
end